/*
 *  Send File program to send stored file on SD card
 *  
 *  Version:                0.7a
 *  Design:                 C. Pham
 *  Implementation:         C. Pham
 *
 *  "T130#" -> Transmit with inter pkt time of 130ms
 *  "Z50#" -> set the pkt size for binary mode, default is 100
 *  "FdesQ5.dat#" -> set the file name to desQ5.dat                
 *  "D0013A2004086D828#" -> set the destination mac addr, use D000000000000FFFF for broadcast again
 *  "I#"/"B#/"S#" -> set to image mode/set to binary mode/set to stream mode
 *  "CR#"/"CS# -> "CR#" reset timer0_millis to reset clock, "CS#" shows current millis() value
 *  "Y0#"/"Y1#" -> disable/enable(set) the 16-bit node's short address
 *  "L0#" -> set flow id to 0x55 in binary mode and to 0x50 in image mode
 *  "R0#"/"R1#" -> disable/enable(set) the MAC layer ack mechanism
 */

/*
 *  Mar, 24th, 2015. v0.7a
 *        the preamble is now 5 bytes long: 0xFF 0x50->0x59 SN Q PKTSIZE
 *  Oct, 30th, 2014. v0.7
 *        improve the support for 125000baud XBee module. At setup, switch automatically to Mac Mode 2 in order to have full 802.15.4 compatibility
 *          uncomment CUSTOM_BAUD_RATE to use an XBee module clocked at 125000
 *        the preamble is now 4 bytes long: 0xFF 0x50->0x59 SN PKTSIZE
 *          PKTSIZE is needed at the receiver to know how many bytes to read from serial port
 *          also, the preamble is the same than the audio board software
 *  Oct, 18th, 2014. v0.6
 *        adds the possibility of toggling between MAC layer ack (XBee MM mode 2) and MAC layer no ack (XBee MM mode 1)
 *        adds flow id in the preamble, 0xFF/0x55-0x59 for binary and 0xFF/0x50-0x54 for image mode. Use command "@/L0-4#" to set the flow id to 
 *        0x55-0x59 in binary mode and to 0x50-0x54 in image mode. This can be used to differenciate flow id at relay nodes.
 *  Dec, 21th, 2013. v0.5
 *        adds framing bytes (preamble), 0xFF0x55 for binary mode and 0xFF0x50 for image mode (XBeeReceive has been updated: use -framing option)
 *  Nov, 4th, 2013. v0.45
 *        adds stream mode, do not insert any header (filesize and offset in file) in binary mode sending 
 *  Jul, 22nd, 2013. v0.4
 *        adds 16-bit reception support  
 *  Jun, 17th, 2013. v0.35
 *        adds 16-bit dest. address support, "D1234" set the dest. address to 1234 and set the XBEE short address to the last 4-digit of the XBee 64-bit MAC address, enabling
 *           (1) other motes working in 16-bit address mode to send packets to the XBee module (especially MicaZ and TelosB with CC2420 radio under TinyOS 
 *           (2) correct 16-bit source addr in the packets that will be sent by the node
 *        adds command "Y0#"/"Y1#" to disable/enable 16-bit node's short address
 *  Jun, 14th, 2013. v0.3
 *        adds command string prefix to '/@". All existing command should be prefixed such as: "/@Z50#"
 *  May 17th, 2013. v0.2
 *      - adds support for DigiMesh interface
 *  April, 6th, 2013. V0.1
 *        
 */

/* TODO
 *        
 */
#include <SoftwareSerial.h>
#include <XBee.h>
#include <Streaming.h> 
#include <SD.h>

// BEGIN of compilation #define statements
//
// read in SD card block by block
#define BLOCK_READ
// print the image data read from SD card
//#define PRINT_IMG_PKT 
// print the binary data read from SD card
//#define PRINT_BIN_PKT
// uses an additional LCD display for more information
#define WITH_LCD
// uses a DigiMesh module for sending packets
//#define WITH_DIGIMESH
// use custom baud rate
#define CUSTOM_BAUDRATE
//
// END of compilation #define statements

#define DEFAULT_IMG_FILE         "japanQ20.dat"
//#define DEFAULT_IMG_FILE         "8kgsm.wav"
//#define DEFAULT_IMG_FILE         "test2400.bit"
//#define DEFAULT_IMG_FILE         "test8000.spx"
//#define PERIODIC_SEND            5000
#define PREAMBLE_SIZE            5
#define MAX_PKT_SIZE             100-PREAMBLE_SIZE // in bytes, all the XBee radio payload
#define MIN_PKT_SIZE             10 // in bytes, for binary, we have 4 extra bytes so the real payload is 10-4=6 which is the minimum size of codec2 audio
#define DEFAULT_INTER_PKT_TIME   130
#define MIN_INTER_PKT_TIME       10 // in ms
#define BROADCAST_MAC_ADDR       "000000000000FFFF"
#define SHORT_BROADCAST_MAC_ADDR "FFFF"
#define FINAL_DEST_MAC_ADDR      "0013A2004086D82E"
//#define FINAL_DEST_MAC_ADDR      "0013A200406FB3CA"  // R1 Libelium home
//#define FINAL_DEST_MAC_ADDR      "0013A200408BC823"  // final rcv home
//#define FINAL_DEST_MAC_ADDR      "0070"  // 16-bit short address, typically a TelosB or MicaZ under TinyOS
//#define FINAL_DEST_MAC_ADDR      "0013A200406FB3E9"  // R1 Libelium office
//#define FINAL_DEST_MAC_ADDR      "0013A200408BC81F"  // final rcv office

#define SHORT_ADDR_FROM_64MAC 1
#define SHORT_ADDR_FFFF 2

// BEWARE, we limit the size of the received cmd to avoid strange packets arriving. 
// So avoid to have to long cmd, use multiple cmd is necessary
#define MAX_CMD_LENGTH           60

// maximum of flow id 0->4
#define MAX_FLOW_ID              4

#ifdef CUSTOM_BAUDRATE
// can use 125000 for faster UART operation but need an XBee module configure for this baud rate
#define MY_XBEE_BAUDRATE         125000
#else
#define MY_XBEE_BAUDRATE         38400 
#endif

XBee xbee = XBee();

// useful AT commands
// get serial high
uint8_t shCmd[] = {'S','H'};
// get serial low
uint8_t slCmd[] = {'S','L'};
// get ack failures low
uint8_t eaCmd[] = {'E','A'};
// get cca failures low
uint8_t ecCmd[] = {'E','C'};
// get/set mac mode
uint8_t mmCmd[] = {'M','M'};
// get/set MY short address
uint8_t myCmd[] = {'M','Y'};

AtCommandRequest atRequest = AtCommandRequest();
AtCommandResponse atResponse = AtCommandResponse();

XBeeResponse response = XBeeResponse();

// create reusable response objects for responses we expect to handle 
#ifdef WITH_DIGIMESH
DMRxResponse rx64 = DMRxResponse();
DMTxStatusResponse txStatus = DMTxStatusResponse();
#else
Rx64Response rx64 = Rx64Response();
Rx16Response rx16 = Rx16Response();
TxStatusResponse txStatus = TxStatusResponse();
#endif

XBeeAddress64 destAddr64;
uint16_t destAddr16;

uint8_t option = 0;
uint8_t* rcv_data = NULL;
uint8_t* rcv_rawData = NULL;
uint8_t rcv_data_len=0;

String macAddr="";

// to hold the data to be sent
char myBuff[MAX_PKT_SIZE+20];

// multi-purpose buffer for various information display
char banner[100];
// used mainly for the LCD display, typically remind the packet size and the generation frequency
char TGInfoStr[80];

// will wrap at 255
uint8_t nbSentPackets=0;
uint8_t chunkSize=MAX_PKT_SIZE;
unsigned long lastSentTime=0;
unsigned long lastImmediateSendTime=0;
unsigned long startGetDataTime=0;
int nbSendRemain=1;
int myMacMode=-1;
long inter_binary_pkt=DEFAULT_INTER_PKT_TIME;

// need to reserve 1 byte more for the NULL-terminated character
// broadcast by default
//char destMACAddr[17]=BROADCAST_MAC_ADDR; 
char destMACAddr[17]=FINAL_DEST_MAC_ADDR;

File myFile;
File root;

char readStr[5];
char theFilename[40] = DEFAULT_IMG_FILE;

boolean image_mode=true;
boolean binary_mode=false;
boolean stream_mode=false;
boolean have_short_address=false;
// default behavior is to use framing bytes
boolean with_framing_bytes=true;

// use flow id of image mode 0x50
uint8_t pktPreamble[PREAMBLE_SIZE] = {0xFF,0x50,0x00,0x00,0x00};
uint8_t flowId = 0x00;

unsigned long QualityFactor;

// the counter to count the number of millis on Arduino
extern volatile unsigned long timer0_millis;

#ifdef WITH_LCD
#include <Wire.h>
#include <LiquidCrystal_I2C.h>

LiquidCrystal_I2C lcd(0x20,16,2);  // set the LCD address to 0x20 for a 16 chars and 2 line display
#endif

void getTGInfo(int index=0) {

        switch (index) {
          
          case 0:
              sprintf(TGInfoStr,"F%s#%c#", theFilename, image_mode?'I':(stream_mode?'S':'B')); 
              break;
              
          case 1:
        
              sprintf(TGInfoStr,"L%.2X#", (image_mode?0x50:0x55)+flowId);
              break;
              
          default:
              break;
        }                     
}

#ifdef WITH_LCD

void printLCD(int theLine, const char* theInfo, boolean clearFirst=true) {
        if (clearFirst)
              lcd.clear();
        lcd.setCursor(0, theLine);
        lcd.print(theInfo);
}

void printInfoLCD(int index=0) {
        lcd.clear();
        getTGInfo(index);
        lcd.home();
        lcd.print(TGInfoStr);
        lcd.setCursor(0,1);
        printLCD(1, destMACAddr, false);
}

#endif

int freeMemory () {
  extern int __heap_start, *__brkval; 
  int v; 
  return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval); 
}

uint16_t read4char() {
      
        for (int i=0; i<4; i++)
            readStr[i]=myFile.read();
        readStr[4]='\0';    
        // skip the space    
        myFile.read();
}

uint8_t read2char() {

        for (int i=0; i<2; i++)
            readStr[i]=myFile.read();
        readStr[2]='\0';    
        // skip the space    
        myFile.read();          
}

int openFile(char* theFile) {
  
        // Check to see if the file exists:
        if (SD.exists(theFile)) {
                Serial.print(theFile);
                Serial.println(" exists.");
        }
        else {
                Serial.print(theFile);
                Serial.println(" doesn't exist.");
                return 1;
        }
 
        // re-open the file for reading:
        myFile = SD.open(theFile);
        
        if (myFile) {
                Serial.print(theFile);
                Serial.println(":");
        }
        else {
                // if the file didn't open, print an error:
                Serial.print("error opening ");
                Serial.println(theFile);
                return 1;
        }          

        return 0;
}        

void prepareImgFile() {
  
        // read the header, 4x16-bit
        for (int i=0; i<4; i++) {
              read4char();
              Serial.print(readStr);
              Serial.print(" ");
        }
        Serial.println("");            
        
        // get the Quality Factor, the last read bytes
        QualityFactor=strtoul(readStr, NULL, 16);
}

uint8_t getNextBinaryPkt(uint8_t* buff, uint8_t s) {

        startGetDataTime = millis();
        
        if (myFile.available()) {
              
              int readVal;
              
              readVal=myFile.read(buff, s);
              
              return readVal;
        }  

        return 0;  
}

uint8_t getNextImgPkt(uint8_t* buff) {
          
        unsigned long nbByte;    

        startGetDataTime = millis();
        
        if (myFile.available()) {
              
              read4char();
              nbByte=strtoul(readStr, NULL, 16);
              Serial.print(nbByte, HEX);
              Serial.println(":");
        
              long startReadPktTime=millis();

              unsigned long v;
#ifdef BLOCK_READ
              int i=0;
              uint8_t imgPkt[300];
              
              // reading by block is much faster, about 2-3ms for 100 bytes
              myFile.read(imgPkt, nbByte*3);
              
              //imgPkt[nbByte*3-1]='\0';
              //Serial.println((char*)imgPkt);
              
              for (int k=1; k<=nbByte; k++) {
                    // set the null character to separate each byte in str format
                    imgPkt[k*3-1]='\0';
                    v=strtoul((char*)&imgPkt[(k-1)*3], NULL, 16);
                    buff[i]=(uint8_t)v;
#ifdef PRINT_IMG_PKT                            
                    Serial.print(buff[i], HEX);
                    Serial.print(" ");
#endif    
                    i++;
              }

#else       

              for (int i=0; i<nbByte; i++) {
                    read2char();
                    v=strtoul(readStr, NULL, 16);
                    buff[i]=(uint8_t)v;
#ifdef PRINT_IMG_PKT                            
                    Serial.print(buff[i], HEX);
                    Serial.print(" ");
#endif                            
              }
#endif                                            
              Serial.print("\nRead in ");
              Serial.print(millis()-startReadPktTime);
              Serial.println(" ms.");
              
              return (int)nbByte;
        }
        else
                return 0;
}

int sendPkt(uint8_t* buff, uint8_t blen) {
                      
        long startSendTime, stopSendTime, previousLastSendTime;
        
#ifdef WITH_DIGIMESH
        // option is 0x00
        DMTxRequest tx = DMTxRequest(destAddr64, DM_BROADCAST_RADIUS_MAX_HOPS, 0x00, buff, blen, nbSentPackets);
#else
        startSendTime=millis();
        previousLastSendTime=lastSentTime;
        
        // we set lastSendTime before the call to the sending procedure in order to be able to send packets back to back since 
        // the sending procedure can introduce a delay 
        lastSentTime=(unsigned long)startSendTime;
        
        // 16-bit destination address?        
        if (strlen(destMACAddr)==4) {
                // set the frame id to the pkt sequence number to keep track of ACK failures when mac mode is 0/2 with unicast address
                Tx16Request tx = Tx16Request(destAddr16, ACK_OPTION, buff, blen, nbSentPackets);
        
                // relay the frame
                xbee.send(tx);              
        }
        else {
                // set the frame id to the pkt sequence number to keep track of ACK failures when mac mode is 0/2 with unicast address
                Tx64Request tx = Tx64Request(destAddr64, ACK_OPTION, buff, blen, nbSentPackets);
        
                // relay the frame
                xbee.send(tx);          
        }
#endif

        stopSendTime=millis();            
        
        nbSentPackets++;
        Serial.print("\nPkt sent ");
        Serial.print(startSendTime);
        
        // read incoming pkt, to get the transmit status, for 5ms max
        xbee.readPacket(5);
        
        if (xbee.getResponse().isAvailable()) {
  
#ifdef WITH_DIGIMESH
               // is it a response to the previously sent packet?            	
  	      if (xbee.getResponse().getApiId() == DM_TX_STATUS_RESPONSE) {
  
  	            xbee.getResponse().getDMTxStatusResponse(txStatus);	
#else          
              // is it a response to the previously sent packet?            	
	      if (xbee.getResponse().getApiId() == TX_STATUS_RESPONSE) {
                      
	            xbee.getResponse().getTxStatusResponse(txStatus);	
#endif   

                    if (txStatus.isSuccess()) {
                          Serial.print(" Success ");        
                    }                        
                    else {                     
                          Serial.print(" Error: 0x");
#ifndef WITH_DIGIMESH                          
                          Serial.print(txStatus.getStatus(), HEX);
                          Serial.print(". Pkt ");
                          // get back the frame id
                          Serial.print(txStatus.getFrameData()[0]);
#endif
                          Serial.print(". ");     
                    }
              }
              else
                    Serial.print(" Not a TX status ");  
        }
        else
              Serial.print(" Timeout ");
              
        Serial.print(startSendTime-previousLastSendTime);
        Serial.print(" ");
        //Serial.print(". In ");
        Serial.print(stopSendTime-startSendTime);
        //Serial.print("ms. ");               
}
          

long getCmdValue(int &i, char* strBuff=NULL) {
        
        int j=0;
        // a maximum of 16 characters plus the null-terminited character, can store a 64-bit MAC address
        char seqStr[17]="****************";

        // character '#' will indicate end of cmd value
        while ((char)rcv_data[i]!='#' && (i < rx64.getDataLength()) && j<strlen(seqStr)) {
                seqStr[j]=(char)rcv_data[i];
                i++;
                j++;
        }
        
        // put the null character at the end
        seqStr[j]='\0';
        
        if (strBuff) {
                strcpy(strBuff, seqStr);        
        }
        else
                return (atol(seqStr));
}    

void setAddresses() {
        boolean setShortAddress=false;
        
        // 16-bit destination address
        if (strlen(destMACAddr)==4) {
                // to indicate that we need to set a short address for the node because 16-bit destination needs a node's short address as well
                if (have_short_address==false)
                        setShortAddress=true;
                
                if (strcmp(destMACAddr, SHORT_BROADCAST_MAC_ADDR)==0) {
                        // 16-bit addressing: This is the SH + SL address of remote XBee
                        // unless you have MY on the receiving radio set to FFFF, this will be received as a RX16 packet
                        destAddr16 = 0xFFFF;    
                } 
                else {
                        destAddr16=(uint16_t)strtol(destMACAddr, NULL, 16);
                }                                 
        }
        else {
                // we can have a 64-bit destination address and have a short address to enable 16-bit address addressing
                // this is why we do not set haveShortAddress to false
                if (strcmp(destMACAddr, BROADCAST_MAC_ADDR)==0) {
                        // 64-bit addressing: This is the SH + SL address of remote XBee
                        // unless you have MY on the receiving radio set to FFFF, this will be received as a RX16 packet
                        destAddr64 = XBeeAddress64(0x00000000, 0x0000FFFF);   
                }
                else {
                        // need to transform the MAC addr string into 2 uint32_t
                        char myMacHigh[9], myMacLow[9];
                        
                        strncpy(myMacHigh, destMACAddr, 8);
                        strncpy(myMacLow, destMACAddr+8, 8);
                        myMacHigh[8]='\0';
                        myMacLow[8]='\0';
                        
                        destAddr64 = XBeeAddress64((uint32_t)strtoul(myMacHigh, NULL, 16), (uint32_t)strtoul(myMacLow, NULL, 16));     
                }           
        }              
#ifndef WITH_DIGIMESH
        if (setShortAddress) 
                setMYAddr(SHORT_ADDR_FROM_64MAC);  
#endif                                                          
}

void setup() {

        long SendFileStartupTime=millis();
       
        Serial.begin(38400);
        Serial.print("Send File starts at ");   
        Serial.println(SendFileStartupTime);
        
        Serial.println("Init Send File. ");
        Serial.print(freeMemory());
        Serial.println(" bytes of free memory.");

#ifdef WITH_LCD
        lcd.init();
        lcd.backlight();
        lcd.home();
        lcd.print("Send File");      
#endif        
        
        delay(5000);
        Serial.print("\nInitializing SD card...");

#ifdef WITH_LCD
        printLCD(1, "InitSD", false);
#endif
         
        if (SD.begin(53)) {
                Serial.println("initialization done.");
                Serial.println("List of files:");
            
                root = SD.open("/");
                printDirectory(root, 0);
                Serial.println("done!");
                root.close();          
        }
        else {
                Serial.println("initialization failed!");
        } 
        
        // need to pause because there may be conflict with the serial port
        delay(5000);

#ifdef WITH_DIGIMESH
        Serial.println("Init XBee DigiMesh");
#ifdef WITH_LCD
        printLCD(1, "InitDM", false);
#endif
#else
        Serial.println("Init XBee 802.15.4"); 
#ifdef WITH_LCD
        printLCD(1, "Init802", false);
#endif        
#endif  
        // if the baud rate is non standard, output will not be available with the serial monitor
        xbee.begin(MY_XBEE_BAUDRATE);
        delay(1000);
        
        // set command to SH to get high part of MAC address
        atRequest.setCommand(shCmd);  
        while (!sendAtCommand()) {Serial.println("retry command ATSH");};
 
        char adr_part[8];

        // get each value, convert it to string and append to the adr
        for (int i = 0; i < atResponse.getValueLength(); i++) {
              sprintf(adr_part, "%lX", (unsigned long)(atResponse.getValue()[i]));
    
              if (atResponse.getValue()[i]<=0xF) 
                macAddr += 0;
        
              macAddr += adr_part;
        } 
        
        // set command to SL to get low part of MAC address
        atRequest.setCommand(slCmd);  
        while (!sendAtCommand()) {Serial.println("retry command ATSL");};
  
        for (int i = 0; i < atResponse.getValueLength(); i++) {
              sprintf(adr_part, "%lX", (unsigned long)(atResponse.getValue()[i]));
    
              if (atResponse.getValue()[i]<=0xF) 
              macAddr += 0;
        
              macAddr += adr_part;    
        } 
#ifdef CUSTOM_BAUDRATE        
        // if we have an xbee module at 125000, it means that it has mac mode 0 for remote AT command
        // then switch it to MM2. Note that this is not definitive since ATWR is not issued
        setMMMode(2);
#ifdef WITH_LCD
        printLCD(1, "Set XBee in MM2 ", false);
        delay(1000);
#endif         
#endif        
        
#ifndef WITH_DIGIMESH        
        // set command to MM to get the XBee mac mode
        atRequest.setCommand(mmCmd);
        
        int counter = 4;
        while (!sendAtCommand() && counter > 0) {
              Serial.println("retry command ATMM");
              counter--;
        }
        
        if (counter > 0) {
                myMacMode=(int)atResponse.getValue()[0];
        }
        else
                Serial.print("ERROR AT+MM");

        sprintf(banner,"-mac:%s WAITING for command from 802.15.4 interface. XBee mac mode %d", (const char*)&macAddr[0], myMacMode);
#else
        sprintf(banner,"-mac:%s WAITING for command from DigiMesh interface.", (const char*)&macAddr[0]);
#endif
        Serial.println(banner);        
        Serial.println("Wait for command \"FdesQ5.dat#I#D0013A20040762053#T60#\" to send image file desQ5.dat with an inter-pkt time of 60ms to 0013A20040762053");
        Serial.print("Current file: ");
        Serial.println(theFilename);
        Serial.print("Current destination: ");
        Serial.println(destMACAddr);
        Serial.print("Current mode: ");
        Serial.println(image_mode?"Image Mode":"Binary Mode");  

        // set the destination addr and set the node's 16-bit addr if the destination addr is also in 16-bit format  
        setAddresses();  
        // start counting time for sending automatically an image from this point
        lastImmediateSendTime = millis();    
        
#ifdef WITH_LCD
        printLCD(1, (const char*)&macAddr[0], false);
        delay(2000);
        
        printInfoLCD();
#endif  
}

void setMYAddr(int addrMode) {

        // set command to MY to set 16-bit node's short address
        atRequest.setCommand(myCmd); 
        uint16_t myShortAddr=0xFFFF;    
        
        Serial.print("Set node's short 16-bit short address to 0x");
        
        // we set the 16-bit addr by taking the last 4-digit of the 64-bit MAC address
        if (addrMode==SHORT_ADDR_FROM_64MAC) {

                Serial.println((const char*)&macAddr[12]);

                myShortAddr=(uint16_t)strtol((const char*)&macAddr[12], NULL, 16);     
                  
                uint8_t tmpValue;
                
                // need to swap byte order
                tmpValue = *((uint8_t*)&myShortAddr);
                *((uint8_t*)&myShortAddr) = *((uint8_t*)&myShortAddr+1);
                *((uint8_t*)&myShortAddr+1) = tmpValue;
                
                have_short_address = true;
        }
        else {   
                // otherwise, set to 0xFFFF to disable 16-bit address packet reception 
                Serial.print(SHORT_BROADCAST_MAC_ADDR);
                have_short_address = false;
        }
                
        atRequest.setCommandValue((uint8_t*)&myShortAddr);  
        
        atRequest.setCommandValueLength(sizeof(myShortAddr)); 

        while (!sendAtCommand()) {
                Serial.println("retry set command ATMY");
                delay(500);
        }
        
        atRequest.clearCommandValue();

        // set command to WR to write in EEPROM
        //atRequest.setCommand(wrCmd); 
        
        //while (!sendAtCommand()) {
        //        Serial.println("retry command ATWR");
        //        delay(500);
        //}
        
        // set command to read back MY 16-bit node's short address
        atRequest.setCommand(myCmd);    
        while (!sendAtCommand()) {
                Serial.println("retry read command ATMY");
                delay(500);
        }
        
        char my_part[4];
        String myAddr="";

        // get each value, convert it to string and append to the MY addr
        for (int i = 0; i < atResponse.getValueLength(); i++) {
              sprintf(my_part, "%lX", (unsigned long)(atResponse.getValue()[i]));
    
              if (atResponse.getValue()[i]<=0xF) 
                myAddr += 0;
        
              myAddr += my_part;
        }                                                         
        
        Serial.print("Node's short address is now: ");
        Serial.println((char*)&myAddr[0]);                        
}

void setMMMode(uint8_t mmMode) {

        // set command to MY to set 16-bit node's short address
        atRequest.setCommand(mmCmd);     
        uint8_t myMMMode=mmMode;
        
        Serial.print("Set MM mode to ");
        Serial.println(mmMode);
                
        atRequest.setCommandValue((uint8_t*)&myMMMode);  
        
        atRequest.setCommandValueLength(sizeof(myMMMode)); 
        
        while (!sendAtCommand()) {
                Serial.println("retry set command ATMM");
                delay(500);
        }
        
        atRequest.clearCommandValue();
  
        // set command to MM to get the XBee mac mode
        atRequest.setCommand(mmCmd);
        
        int counter = 4;
        while (!sendAtCommand() && counter > 0) {
              Serial.println("retry command ATMM");
              counter--;
        }
        
        if (counter > 0) {
                myMacMode=(int)atResponse.getValue()[0];
        }
        else
                Serial.print("ERROR AT+MM");            
        
        Serial.print("MAC mode is now: ");
        Serial.println(myMacMode);                        
}

void loop()
{
        // read incoming pkt, to get cmd
        xbee.readPacket();

        if (xbee.getResponse().isAvailable()) {

#ifdef WITH_DIGIMESH
                if (xbee.getResponse().getApiId() == DM_RX_RESPONSE) {

                        xbee.getResponse().getDMRxResponse(rx64);
#else
                if (xbee.getResponse().getApiId() == RX_64_RESPONSE || xbee.getResponse().getApiId() == RX_16_RESPONSE) {

                        if (xbee.getResponse().getApiId() == RX_64_RESPONSE) {
                          
                                xbee.getResponse().getRx64Response(rx64);
                                option = rx64.getOption();
                                rcv_rawData = rx64.getFrameData();
                                rcv_data = rx64.getData();
                                rcv_data_len = rx64.getDataLength();
                        }
                        
                        if (xbee.getResponse().getApiId() == RX_16_RESPONSE) {
                          
                                xbee.getResponse().getRx16Response(rx16);
                                option = rx16.getOption();
                                rcv_rawData = rx16.getFrameData();
                                rcv_data = rx16.getData();
                                rcv_data_len = rx16.getDataLength();
                        }                        
#endif   

#ifdef WITH_DIGIMESH
                        Serial.print("RCVCMDDM:");  
#else
                        Serial.print("RCVCMD802:");    
#endif   
                       
                        for (int f=0; f < rcv_data_len; f++) {
                                Serial.print((char)rcv_data[f]);
                        }

                                                
                        // test whether the packet comes from a WaspMote with the Libelium API
                        // be careful here, if it is a fragmented message from WaspMote, we may have problem identifyng it!
                        if ((char)rcv_data[2]=='#' && (char)rcv_data[4]=='W' && (char)rcv_data[5]=='A' && (char)rcv_data[6]=='S' && (char)rcv_data[7]=='P') {
                                // we don't take into account these pkt, in this way, we can control Arduino and WaspMote TG independently
                                Serial.println("RCV from WaspMote, ignoring");
                        }
                        else {
                          
                                int i=0;                        
                                int cmdValue;
                                long longCmdValue;
                                
                                // we use prefix '/@' to indicate a command string
                                if (rcv_data[i]=='/' && rcv_data[i+1]=='@') {
                                
                                        Serial.println("\nParsing command string");
                                        
                                        i=2;
                                
                                        // parse the received command
                                        // typical cmd is "FdesQ5.dat#I#D0013A20040762053#T130#" for instance
                                        while (i < rcv_data_len && rcv_data_len<MAX_CMD_LENGTH) {
#ifdef WITH_LCD                                                
                                                printLCD(1, (char*)(rcv_data+2), true);
                                                delay(1000);
#endif                                                
                                                switch ((char)rcv_data[i]) {
                                                  
                                                // binary packets stream indicator?
                                                // "FdesQ5.dat#I/B#D0013A20040762053#T130#" 
                                                
                                                // T130# -> start sending with inter pkt time of 130ms
                                                case 'T': {
                                                                        
                                                        i++;
                                                        longCmdValue=getCmdValue(i);
                                                        // cannot set an inter pkt time smaller than MIN_INTER_PKT_TIME
                                                        if (longCmdValue < MIN_INTER_PKT_TIME)
                                                                longCmdValue = MIN_INTER_PKT_TIME;
                                                        // set new inter packet time        
                                                        inter_binary_pkt=longCmdValue; 
                                                        
                                                        immediateSendFile(2000,10000);                               
                                                }      
                                                break;                          
                
                                                // set the pkt size for binary mode, default is 100
                                                // "Z50#"
                                                case 'Z':
                                                
                                                        i++;
                                                        cmdValue=getCmdValue(i);
                                                        // cannot set pkt size greater than MAX_PKT_SIZE
                                                        if (cmdValue > MAX_PKT_SIZE)
                                                                cmdValue = MAX_PKT_SIZE;
                                                        // cannot set pkt size smaller than MAX_PKT_SIZE
                                                        if (cmdValue < MIN_PKT_SIZE)
                                                                cmdValue = MIN_PKT_SIZE;
                                                        // set new pkt size        
                                                        chunkSize=cmdValue; 
                                                        
                                                        Serial.print("Set pkt size to ");
                                                        Serial.print(chunkSize);
                                                break;
                
                                                // for clock synchronization
                                                // "CR#"/"CS#"
                                                case 'C':
                                                        if ((char)rcv_data[i+1]=='R') {
                                                                long old_millis=millis();
                                                                timer0_millis=0;
                                                                delay(1000);
                                                                long new_millis=millis();
                                                                Serial.print("Old millis(): ");
                                                                Serial.print(old_millis);
                                                                Serial.print(" Clock reset, new millis() after 1000ms ");
                                                                Serial.println(new_millis);
                                                                
                                                                lastSentTime=0;
                                                                i++;                     
                                                        }
                                                        
                                                        if ((char)rcv_data[i+1]=='S') {
                                                                long current_millis=millis();
                                                                Serial.print("Clock show, current millis(): ");
                                                                Serial.println(current_millis);
                                                                i++;
                                                        }
                                                        
                                                        // skip everything until the '#' character
                                                        i++;
                                                        while ((char)rcv_data[i]!='#' && (i < rcv_data_len))
                                                                i++;                                        
                                                        break;   
                                                        
                                                // FdesQ5.dat -> set the file name to desQ5.dat        
                                                case 'F':
                                                        i++;
                                                        cmdValue=getCmdValue(i, theFilename);
                                                        
                                                        Serial.print("Will open ");
                                                        Serial.println(theFilename);                                                        
                                                        break;                                               
                                                                
                                                // command to set the destination, override the default destination "D0013A20040762053#" will send to 0013A20040762053 
                                                // use D000000000000FFFF for broadcast again 
                                                case 'D':  
                                                                     
                                                      if (  ((char)rcv_data[i+1] >= '0' && (char)rcv_data[i+1] <= '9') || ((char)rcv_data[i+1] >= 'A' && (char)rcv_data[i+1] <= 'F')) {
                                                              Serial.print("Set dest MAC addr to ");    
              
                                                              i++;
                                                              cmdValue=getCmdValue(i, destMACAddr);
                                                              
#ifdef WITH_DIGIMESH
                                                              // DigiMesh does not support 16-bit address mode
                                                              if (strlen(destMACAddr)!=16) {
                                                                      Serial.print("bad address, need 16 bytes in hex format. Restore default destination ");  
                                                                      strcpy(destMACAddr, FINAL_DEST_MAC_ADDR);
                                                              }
#else
                                                              if (strlen(destMACAddr)!=16 && strlen(destMACAddr)!=4) {
                                                                      Serial.print("bad address, need 16 or 4 bytes in hex format. Restore default destination ");  
                                                                      strcpy(destMACAddr, FINAL_DEST_MAC_ADDR);
                                                              }
#endif                                                                 
                                                                  
                                                              Serial.println(destMACAddr);
                                                              
                                                              setAddresses();
                                                      }        
                                                      break;
      
#ifndef WITH_DIGIMESH
                                                // command to set the node's short address to the last 4 digit of the 64-bit MAC address, "Y0#" or "Y1#" 
                                                case 'Y':  
                                                      
                                                      if ((char)rcv_data[i+1]=='0')
                                                              setMYAddr(SHORT_ADDR_FFFF);
                                                      else 
                                                              setMYAddr(SHORT_ADDR_FROM_64MAC);                                                     
                                                                                 
                                                      break;
#endif                                                                                                                
                                                case 'B':
                                                        
                                                        binary_mode=true;
                                                        stream_mode=false;
                                                        image_mode=false;
                                                        // change the framing bytes to 0xFF0x55
                                                        pktPreamble[1]=0x55+flowId;
                                                        
                                                        Serial.println("Set binary mode on");
                                                        
                                                        // skip everything until the '#' character
                                                        i++;
                                                        while ((char)rcv_data[i]!='#' && (i < rcv_data_len))
                                                                i++;                                        
                                                        break;                                                

                                                case 'S':
                                                        
                                                        stream_mode=true;
                                                        binary_mode=true;
                                                        image_mode=false;
                                                        Serial.println("Set stream (binary) mode on");
                                                        
                                                        // skip everything until the '#' character
                                                        i++;
                                                        while ((char)rcv_data[i]!='#' && (i < rcv_data_len))
                                                                i++;                                        
                                                        break;                                                
                                                                                                                                
                                                                
                                                case 'I':
                                                        
                                                        image_mode=true;
                                                        binary_mode=false;
                                                        stream_mode=false;
                                                        // change the framing bytes to 0xFF0x50
                                                        pktPreamble[1]=0x50+flowId;                                                        
                                                        Serial.println("Set image mode on");
                                                        
                                                        // skip everything until the '#' character
                                                        i++;
                                                        while ((char)rcv_data[i]!='#' && (i < rcv_data_len))
                                                                i++;
                                                        break;
                                                        
                                                case 'L':
                                                
                                                        i++;
                                                        cmdValue=getCmdValue(i);
                                                        // cannot set flow id greater than MAX_FLOW_ID
                                                        if (cmdValue > MAX_FLOW_ID)
                                                                cmdValue = 0;
                                                        // set new pkt size        
                                                        flowId=cmdValue; 
                                                        
                                                        Serial.print("Set flow id to ");
                                                        Serial.print(flowId);
#ifdef WITH_LCD                                                        
                                                        printInfoLCD(1);
#endif 
                                                        delay(1000);
                                                        break;          
                                                
                                                case 'R':
                                                        i++;
                                                        cmdValue=getCmdValue(i);
                                                        
                                                        if (cmdValue==0)
                                                              // MAC mode 1, without ack
                                                              setMMMode(1);
                                                        else 
                                                              // MAC mode 2, with ack
                                                              setMMMode(2);                                                                                                                                      
                                                        break;                                               
                
                                                default:
                
                                                        Serial.println("Invalid cmd, skiping");
                                                        // skip everything until the '#' character
                                                        i++;
                                                        while ((char)rcv_data[i]!='#' && (i < rcv_data_len))
                                                                i++;                                                    
                                                                 
                                                        break;
                                                }
                                                
                                                // go to next command if any.
                                                i++;                                    
                                        }
                                }
                                else {
                                        Serial.println("\nHave not recognized a command prefix /@");
                                }
                        }
                        
#ifdef WITH_LCD         
                        printInfoLCD();
#endif     
                }
        }       
     
#ifdef PERIODIC_SEND
        
        if (millis()-lastImmediateSendTime > PERIODIC_SEND && nbSendRemain > 0) {
                immediateSendFile(0,3000);
                lastImmediateSendTime=millis();
                nbSendRemain--;  
        }
#endif

}

void immediateSendFile(int delay_before_send, int delay_after_send) {

        uint8_t nbBinaryPkt=0;
        long totalAppDataSize=0L;
        uint8_t dataSize=0;
        // just the maximum pkt size plus some more bytes
        uint8_t myBuff[110];
        
        Serial.print("Immediate send file with inter-pkt ");
        Serial.print(inter_binary_pkt);
        Serial.print(" to ");
        Serial.print(destMACAddr);
        Serial.print(" in ");
        Serial.print(delay_before_send);
        Serial.println("ms");
        delay(delay_before_send);        
        
        int openStatus=openFile(theFilename);
        
        if (!openStatus) {
                
                long startSendFile, stopSendFile, now;
                uint8_t nextPktSize;
                
                if (with_framing_bytes) {
                        myBuff[0]=pktPreamble[0];
                        myBuff[1]=(image_mode?0x50:0x55)+flowId;
                        dataSize += sizeof(pktPreamble);
                }
                
                if (image_mode) {
#ifdef WITH_LCD
                        printLCD(1, "Send Image      ", false);
#endif 
                        // read the header
                        prepareImgFile();
                              
                        startSendFile=millis();
                        
                        Serial.print("Send Image ");
                        Serial.println(startSendFile);
                        
                        while ( (nextPktSize=getNextImgPkt(myBuff+dataSize)) != 0 ) {
                         
                                dataSize += nextPktSize;
                                
                                // set the sequence number
                                myBuff[2] = nbBinaryPkt;
                                
                                // set the Quality Factor
                                myBuff[3] = (uint8_t)QualityFactor;

                                // set the packet size
                                myBuff[4] = nextPktSize;
                                
                                sendPkt(myBuff, dataSize);
                                
                                // count the preamble if any
                                totalAppDataSize+=dataSize;
                                
                                nbBinaryPkt++;
                                
                                if (with_framing_bytes)
                                        dataSize = sizeof(pktPreamble);
                                else
                                        dataSize = 0;
               
                                while ( (millis()-startGetDataTime) < inter_binary_pkt)
                                       ;                           
                        }
                }
                
                if (binary_mode) {
                  
#ifdef WITH_LCD
                        printLCD(1, "Send Binary     ", false);
#endif                                                   
                        uint16_t offset=0;
                        uint16_t fileSize=0;
                        // we reserve the first bytes to put the filesize and the offset in the binary file
                        uint8_t reserved_bytes = 4;
                        
                        fileSize=myFile.size();
                        Serial.print("File size is ");
                        Serial.println(fileSize);
                                     
                        startSendFile=millis();
                        
                        Serial.print("Send Binary ");
                        Serial.println(startSendFile);  
                        
                        if (stream_mode) {
                                // do not insert anything in stream mode, just the framing bytes (if any) and payload
                                reserved_bytes=0;        
                        }
                        
                        // we reserve the first 4 bytes to put the filesize and the offset in the binary file
                        // when indicating a chunksize, you may have to add the 4 extra reserved bytes, chunksize=10 means that you will put 6 bytes in the payload
                        while ( (nextPktSize=getNextBinaryPkt(myBuff+dataSize+reserved_bytes, chunkSize-reserved_bytes)) != 0 ) {
                                        
                                if (reserved_bytes==4) {       
                                        // store high part
                                        myBuff[dataSize]=(uint8_t) (fileSize >> 8);  
                                        // store low part
                                        myBuff[dataSize+1]=(uint8_t) (fileSize & 0x00FF);
                                        
                                        // store high part
                                        myBuff[dataSize+2]=(uint8_t) (offset >> 8);  
                                        // store low part
                                        myBuff[dataSize+3]=(uint8_t) (offset & 0x00FF);
                                }
                                
                                dataSize += nextPktSize;
#ifdef PRINT_BIN_PKT
                                for (int k=0; k<dataSize+reserved_bytes; k++) {
                                        Serial.print(myBuff[k], HEX);
                                        Serial.print(" "); 
                                }                                                     
#endif
                                
                                // set the sequence number
                                myBuff[2] = nbBinaryPkt;
                                // set the total payload size, including the number of reserved bytes
                                myBuff[3] = chunkSize;
                                
                                sendPkt(myBuff, dataSize+reserved_bytes);
                                totalAppDataSize += dataSize+reserved_bytes;
                                offset += nextPktSize;
                                nbBinaryPkt++;

                                if (with_framing_bytes)                                        
                                        dataSize = sizeof(pktPreamble);                                   
                                else
                                        dataSize = 0;                                
                                
                                while ( (millis()-startGetDataTime) < inter_binary_pkt)
                                       ;
                        }
                        
                }
                
                stopSendFile=millis();
                
                Serial.print("Sent ");
                Serial.print(nbBinaryPkt);
                Serial.print(" binary pkt (");
                Serial.print(totalAppDataSize);
                Serial.print("B) in ");
                
                Serial.print(stopSendFile-startSendFile);
                Serial.println(" ms.");
                Serial.print("Sending app throughput is ");
                
                double thr=totalAppDataSize*8/((stopSendFile-startSendFile)/1000);
                
                Serial.print(thr);
                Serial.println("bps");
                
                // close the file:
                myFile.close();
#ifdef WITH_LCD
                sprintf(TGInfoStr,"%d pkt %dms", nbBinaryPkt, stopSendFile-startSendFile);
                printLCD(0, (const char*)TGInfoStr, true);
                sprintf(TGInfoStr,"Th. %d bps", (int)thr);
                printLCD(1, (const char*)TGInfoStr, false);
                delay(delay_after_send);
#endif                                                                                              
        }                
#ifdef WITH_LCD         
        printInfoLCD();
#endif  
}

void printDirectory(File dir, int numTabs) {
  
        while(true) {
                 
                File entry =  dir.openNextFile();
                 
                if (!entry) {
                        // no more files
                        //Serial.println("**nomorefiles**");
                        break;
                }
                 
                for (uint8_t i=0; i<numTabs; i++) {
                        Serial.print('\t');
                }
                 
                Serial.print(entry.name());
                 
                if (entry.isDirectory()) {
                        Serial.println("/");
                        printDirectory(entry, numTabs+1);
                } 
                else {
                        // files have sizes, directories do not
                        Serial.print("\t\t");
                        Serial.println(entry.size(), DEC);
                }
        }
}

// used here to send AT command to the XBee to get information such as the MAC address
int sendAtCommand() {
#ifdef DEBUGPRINT
        Serial.println("Sending command to the Serial");
#endif
        // send the command
        xbee.send(atRequest);

        // wait up to 5 seconds for the status response
        if (xbee.readPacket(5000)) {
                // got a response!
                // should be an AT command response
                if (xbee.getResponse().getApiId() == AT_COMMAND_RESPONSE) {
                        xbee.getResponse().getAtCommandResponse(atResponse);
            
                        if (atResponse.isOk()) {
#ifdef DEBUGPRINT        
                                Serial.print("Command [");
                                Serial.print((char)atResponse.getCommand()[0]);
                                Serial.print((char)atResponse.getCommand()[1]);
                                Serial.println("] was successful!");
#endif
                                if (atResponse.getValueLength() > 0) {
#ifdef DEBUGPRINT          
                                        Serial.print("Command value length is ");
                                        Serial.println(atResponse.getValueLength(), DEC);
      
                                        Serial.print("Command value: 0x");
                                        
                                        for (int i = 0; i < atResponse.getValueLength(); i++) {
                                          Serial.print(atResponse.getValue()[i], HEX);
                                          Serial.print(" ");
                                        }
                              
                                        Serial.println("");
#endif          
                                }
                                return 1;
                        } 
                        else {
                                Serial.print("Command return error code: 0x");
                                Serial.println(atResponse.getStatus(), HEX);
                        }
                } 
                else {
                        Serial.print("Expected AT response but got 0x");
                        Serial.print(xbee.getResponse().getApiId(), HEX);
                }   
                
                
        } 
        else {
              // at command failed
              if (xbee.getResponse().isError()) {
                    Serial.print("Error reading packet.  Error code: 0x");  
                    Serial.println(xbee.getResponse().getErrorCode(), HEX);
              } 
              else {
                    Serial.print("No response from radio. ");  
              }
        }
        return 0;
}
